// Copyright (c) 2023 ETH Zurich.
//                    All rights reserved.
//
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.
//
// main author: Marc Fischer

#include <inttypes.h>
#include <string.h>

#include "gda_list.h"

// TODO: remove this
#ifndef UNUSED
  #define UNUSED(x) (void)(x)
#endif


/**
  Private function to remove elements from a given range.
*/
inline int list_p_free_range(NOD_Node* start, NOD_Node* end) {
  if(start == NULL || end == NULL) {
    return 0;
  }

  int i = 0;
  while(start != end) {
    NOD_Node* tmp_node = start;
    start = start->next;
    free(tmp_node->value);
    free(tmp_node);
    i++;
  }
  free(end->value);
  free(end);
  i++;

  return i;
}


/**
  Private function to create a new node with the given value and element size.
 */
inline NOD_Node* list_p_create_node(void* value, size_t element_size) {
  NOD_Node* node = (NOD_Node*) malloc(sizeof(NOD_Node));
  node->value = (void*)malloc(sizeof(element_size));
  memcpy(node->value, value, element_size);
  return node;
}


/**
  Private function to free a node.
 */
inline void list_p_free_node(NOD_Node* node) {
  if(node) {
    free(node->value);
    free(node);
  }
}


void NOD_list_create(NOD_List** list, int element_size) {
  *list = (NOD_List*) malloc(sizeof(NOD_List));
  (*list)->head = NULL;
  (*list)->tail = NULL;
  (*list)->size = 0;
  (*list)->element_size = element_size;
}


void NOD_list_free(NOD_List** list) {
  NOD_Node* head = (*list)->head;
  NOD_Node* tail = (*list)->tail;

  list_p_free_range(head, tail);

  free(*list);
  *list = NULL;
}


size_t NOD_list_size(NOD_List* list) {
  return list->size;
}


NOD_Node* NOD_list_push_back(NOD_List* list, void* value) {
  NOD_Node* node = list_p_create_node(value, list->element_size);

  if(list->size == 0) {
    list->head = node;
    list->tail = node;
    node->next = NULL;
    node->prev = NULL;
  }
  else {
    node->prev = list->tail;
    list->tail->next = node;
    node->next = NULL;
    list->tail = node;
  }

  list->size++;
  return node;
}


NOD_Node* NOD_list_push_front(NOD_List* list, void* value) {
  NOD_Node* node = list_p_create_node(value, list->element_size);

  if(list->size == 0) {
    list->head = node;
    list->tail = node;
    node->next = NULL;
    node->prev = NULL;
  }
  else {
    node->next = list->head;
    list->head->prev = node;
    node->prev = NULL;
    list->head = node;
  }

  list->size++;

  return node;
}


NOD_Node* NOD_list_insert_before(NOD_List* list, NOD_Node* position, void* value) {
  // we insert at the head
  if(position == list->head) {
    return NOD_list_push_front(list, value);
  }
  else {
    NOD_Node* node = list_p_create_node(value, list->element_size);

    NOD_Node* right = position;
    NOD_Node* left = position->prev;

    left->next = node;
    node->next = right;
    right->prev = node;
    node->prev = left;
    list->size++;

    return node;
  }
}


NOD_Node* NOD_list_insert_after(NOD_List* list, NOD_Node* position, void* value) {
  // we insert at the tail
  if(position == list->tail) {
    return NOD_list_push_back(list, value);
  }
  else {
    NOD_Node* node = list_p_create_node(value, list->element_size);

    NOD_Node* right = position->next;
    NOD_Node* left = position;

    left->next = node;
    node->next = right;
    right->prev = node;
    node->prev = left;
    list->size++;
    return node;
  }
}


void NOD_list_pop_back(NOD_List* list) {
  NOD_Node* tail = list->tail;
  if(list->size == 0) {
    return;
  }
  else if(list->size == 1) {
    list->head = NULL;
    list->tail = NULL;
    list->size = 0;
  }
  else {
    list->tail = list->tail->prev;
    list->tail->next = NULL;
    list->size--;
  }

  list_p_free_node(tail);
}


void NOD_list_pop_front(NOD_List* list) {
  NOD_Node* head = list->head;
  if(list->size == 0) {
    return;
  }
  else if(list->size == 1) {
    list->head = NULL;
    list->tail = NULL;
    list->size = 0;
  }
  else {
    list->head = list->head->next;
    list->head->prev = NULL;
    list->size--;
  }

  list_p_free_node(head);
}


NOD_Node* NOD_list_next(NOD_List* list, NOD_Node* node) {
  UNUSED(list);
  return node->next;
}


NOD_Node* NOD_list_prev(NOD_List* list, NOD_Node* node) {
  UNUSED(list);
  return node->prev;
}


void* NOD_list_value(NOD_List* list, NOD_Node* node) {
  UNUSED(list);
  return node->value;
}


size_t NOD_list_to_array(NOD_List* list, void* array, size_t size_in_elements) {
  NOD_Node* head = list->head;
  size_t element_size = list->element_size;

  size_t i = 0;
  char* char_arr = (char*) array;
  for( ; head != NULL && i < size_in_elements; i++) {
    memcpy(char_arr + i*element_size, head->value, element_size);
    head = head->next;
  }
  return i;
}


NOD_Node* NOD_list_erase_single(NOD_List* list, NOD_Node* node) {
  if(list->head == node) {
    NOD_list_pop_front(list);
    return list->head;
  }
  else if(list->tail == node) {
    NOD_list_pop_back(list);
    return list->tail;
  }
  else {
    NOD_Node* left = node->prev;
    NOD_Node* right = node->next;
    left->next = right;
    right->prev = left;
    list->size--;

    list_p_free_node(node);
    return right;
  }
}


NOD_Node* NOD_list_erase_range(NOD_List* list, NOD_Node* start, NOD_Node* end) {
  if(start == end) {
    return NOD_list_erase_single(list, start);
  }
  else if( list->head == start && list->tail == end ) {
    list_p_free_range(start, end);

    list->head = NULL;
    list->tail = NULL;
    list->size = 0;
    return NULL;
  }
  else if( list->head == start) {
    NOD_Node* end_next = end->next;
    end_next->prev = NULL;
    list->head = end_next;

    int i = list_p_free_range(start, end);
    list->size = list->size - i;
    return list->head;
  }
  else if(list->tail == end) {
    NOD_Node* start_prev = start->prev;
    start_prev->next = NULL;
    list->tail = start_prev;

    int i = list_p_free_range(start, end);
    list->size = list->size - i;
    return list->tail;
  }
  else {
    NOD_Node* start_prev = start->prev;
    NOD_Node* end_next = end->next;
    start_prev->next = end_next;
    end_next->prev = start_prev;

    int i = list_p_free_range(start, end);
    list->size = list->size - i;
    return end_next;
  }
}


void NOD_list_map(NOD_List* list, void (*fun)(void*) ) {
  NOD_Node* head = list->head;
  while(head != NULL) {
    fun( (void*)head);
    head = head->next;
  }
}


NOD_Node* NOD_list_find_node(NOD_List* list, NOD_Node* node) {
  NOD_Node* head = list->head;
  while(head != NULL) {
    if(head == node) {
      return head;
    }
    head = head->next;
  }
  return NULL;
}


NOD_Node* NOD_list_find_value(NOD_List* list, void* value) {
  NOD_Node* head = list->head;
  while(head != NULL) {
    if( memcmp(value, head->value, list->element_size) == 0) {
      return head;
    }
    head = head->next;
  }
  return NULL;
}


NOD_Node* NOD_list_front(NOD_List* list) {
  return list->head;
}


NOD_Node* NOD_list_back(NOD_List* list) {
  return list->tail;
}
