// Copyright (c) 2023 ETH Zurich.
//                    All rights reserved.
//
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.
//
// main author: Marc Fischer

#ifndef __NOD_LIST_H
#define __NOD_LIST_H

#include <stdlib.h>

/**
  This header provides functions for a doubly-linked list.
  The list provides an interface similar to std::list.
  Note that data that is stored in the list gets copied.
  Upon retrieval, a pointer to the memory location is returned,
  where the value is stored.
  It provides the following properties:
    - O(1) insert at the head or tail
    - O(1) removal at the head, tail or when given a pointer
    - O(n) destruction (O(1) amortized)
    - O(n) search
*/


/**
  data type definitions
 */
typedef struct NOD_Node {
  struct NOD_Node* next;
  struct NOD_Node* prev;
  void* value;
} NOD_Node;

typedef struct NOD_List {
  struct NOD_Node* head;
  struct NOD_Node* tail;
  /**
    the number of elements
   */
  size_t size;
  /**
    the size of the value a node holds
   */
  size_t element_size;
} NOD_List;


/**
  function prototypes
 */

/**
  Initializes a new list.
  The list should be deleted by calling list_free.
 */
void NOD_list_create(NOD_List** list, int element_size);

/**
  Removes all elements from list, frees them and frees the list itself.
  The list is set to NULL.
 */
void NOD_list_free(NOD_List** list);

/**
  Returns the size of the list (the number of elements stored in the list);
 */
size_t NOD_list_size(NOD_List* list);

/**
  Adds a new element at the end. The value gets copied.
  The node element that stores the element is returned.
 */
NOD_Node* NOD_list_push_back(NOD_List* list, void* value);

/**
  Adds a new element at the front. The value gets copied.
  The node element that stores the element is returned.
 */
NOD_Node* NOD_list_push_front(NOD_List* list, void* value);

/**
  Inserts a new node before the node called position in the list named list.
  The value gets copied.
  The node element that stores the element is returned.
 */
NOD_Node* NOD_list_insert_before(NOD_List* list, NOD_Node* position, void* value);

/**
  Inserts the node element after the node position in the list named list.
  The node element that stores the element is returned.
*/
NOD_Node* NOD_list_insert_after(NOD_List* list, NOD_Node* position, void* value);

/**
  Delete the last element.
 */
void NOD_list_pop_back(NOD_List* list);

/**
  Deletes the first element.
*/
void NOD_list_pop_front(NOD_List* list);

/**
  Get the next element from the list - if none exists, NULL is returned.
 */
NOD_Node* NOD_list_next(NOD_List* list, NOD_Node* node);

/**
  Get the previous element from the list - if none exists, NULL is returned.
 */
NOD_Node* NOD_list_prev(NOD_List* list, NOD_Node* node);

/**
  Get the value to which the node points to.
 */
void* NOD_list_value(NOD_List* list, NOD_Node* node);

/**
  Copies the values of the list into the given array. Returns the amount of elements
  that are written to the array.
 */
size_t NOD_list_to_array(NOD_List* list, void* array, size_t size_in_elements);

/**
  Erase a single element. The node will be freed.
  Returns the node element that is after the removed element.
  If the head is removed, the new head is returned.
  If the tail is removed, the tail is returned.
  If the list is empty, null is returned.
*/
NOD_Node* NOD_list_erase_single(NOD_List* list, NOD_Node* node);

/**
  Erase multiple elements in a given range, including the given nodes. All affected nodes will be freed.
  TODO: The explanation below needs to be improved, since multiple elements are removed.
  Returns the node elment that is after the removed element.
  If the head is removed, the new head is returned.
  If the tail is removed, the tail is returned.
  If the list is empty, null is returned.
 */
NOD_Node* NOD_list_erase_range(NOD_List* list, NOD_Node* start, NOD_Node* end);

/**
  Apply the given function on every element of the list
 */
void NOD_list_map(NOD_List* list, void (*fun)(void*) );

/**
 Returns a pointer to the element if it is found, and NULL otherwise.
 */
NOD_Node* NOD_list_find_node(NOD_List* list, NOD_Node* node);

/**
  Returns a pointer to the element if it is found, and NULL otherwise.
 */
NOD_Node* NOD_list_find_value(NOD_List* list, void* value);

/**
  Returns the first node (the head).
 */
NOD_Node* NOD_list_front(NOD_List* list);

/**
  Returns the last node (the tail).
 */
NOD_Node* NOD_list_back(NOD_List* list);

#endif // __NOD_LIST_H
